import importlib.metadata
import warnings

from mkdocs_jupyterlite._plugin import JupyterlitePlugin as JupyterlitePlugin
from mkdocs_jupyterlite._plugin import (
    JupyterlitePluginConfig as JupyterlitePluginConfig,
)

try:
    __version__ = importlib.metadata.version(__name__)
except importlib.metadata.PackageNotFoundError as e:
    warnings.warn(f"Could not determine version of {__name__}\n{e!s}", stacklevel=2)
    __version__ = "unknown"



from __future__ import annotations

import logging
import shutil
import tempfile
from collections.abc import Iterable
from pathlib import Path
from typing import Any, Literal

import gitmatch
import markdown
import nbformat
from mkdocs.config.base import Config as BaseConfig
from mkdocs.config.config_options import ListOfItems, SubConfig
from mkdocs.config.config_options import Type as OptionType
from mkdocs.config.defaults import MkDocsConfig
from mkdocs.plugins import BasePlugin
from mkdocs.structure.files import File, Files
from mkdocs.structure.pages import Page
from mkdocs.structure.toc import TableOfContents, get_toc
from nbconvert import MarkdownExporter

from mkdocs_jupyterlite import _build

log = logging.getLogger("mkdocs.plugins.jupyterlite")


class NotebookFile(File):
    """
    Wraps a regular File object to make .ipynb files appear as valid documentation files.
    """

    def __init__(self, file: File) -> None:
        self._file = file

    def __getattr__(self, name: str) -> Any:
        return self._file.__getattribute__(name)

    def is_documentation_page(self) -> Literal[True]:
        return True


class _WheelConfig(BaseConfig):
    command = OptionType(str, default="")
    url = OptionType(str, default="")


class JupyterlitePluginConfig(BaseConfig):
    enabled = OptionType(bool, default=True)
    notebook_patterns = OptionType(list, default=["**/*.ipynb"])
    wheels = ListOfItems(SubConfig(_WheelConfig), default=[])


class JupyterlitePlugin(BasePlugin[JupyterlitePluginConfig]):
    def __init__(self):
        super().__init__()
        if isinstance(self.config, dict):
            plugin_config = JupyterlitePluginConfig()
            plugin_config.load_dict(self.config)
            self.config = plugin_config
        self._jupyterlite_build_dir = tempfile.TemporaryDirectory()

    def _cleanup(self) -> None:
        log.info(
            "[jupyterlite] cleaning up temporary build directory: "
            + str(self._jupyterlite_build_dir.name)
        )
        self._jupyterlite_build_dir.cleanup()

    def on_files(self, files: Files, config: MkDocsConfig) -> Files:
        outfiles = []
        # paths to notebooks relative to config.docs_dir
        notebook_relative_paths = []
        log.info("[jupyterlite] looking for notebook files in " + str(config.docs_dir))
        for file in files:
            if is_notebook(
                relative_path=file.src_uri,
                notebook_patterns=self.config.notebook_patterns,
            ):
                log.info("[jupyterlite] including notebook: " + str(file.src_uri))
                outfiles.append(NotebookFile(file))
                notebook_relative_paths.append(file.src_uri)
            else:
                log.debug("[jupyterlite] ignoring file: " + str(file.src_uri))
                outfiles.append(file)
        _build.build_site(
            docs_dir=Path(config.docs_dir),
            notebook_relative_paths=notebook_relative_paths,
            wheel_sources=self.config.wheels,
            output_dir=Path(self._jupyterlite_build_dir.name),
        )
        return Files(outfiles)

    def on_pre_page(
        self, page: Page, /, *, config: MkDocsConfig, files: Files
    ) -> Page | None:
        if not isinstance(page.file, NotebookFile):
            return page
        iframe_src = f"{config.site_url}jupyterlite/notebooks/index.html?path={page.file.src_uri}"

        def new_render(self: Page, config: MkDocsConfig, files: Files) -> None:
            log.debug("[jupyterlite] rendering " + page.file.abs_src_path)
            log.debug("[jupyterlite] creating iframe with src " + iframe_src)
            body = f"""
            <iframe src="{iframe_src}"
                width="100%"
                height="800px"
                frameborder="1">
            </iframe>
            """
            self.content = body
            toc, title_in_notebook = get_nb_toc_and_title(page.file.abs_src_path)
            log.debug("[jupyterlite] TOC: " + str(toc))
            log.debug("[jupyterlite] title in notebook: " + str(title_in_notebook))
            log.debug("[jupyterlite] page title: " + str(self.title))
            self.toc = toc
            if title_in_notebook and not self.title:
                self.title = title_in_notebook

        # replace render with new_render for this object only
        page.render = new_render.__get__(page, Page)
        return page

    def on_post_build(self, config: MkDocsConfig) -> None:
        shutil.copytree(
            self._jupyterlite_build_dir.name,
            Path(config.site_dir) / "jupyterlite",
            dirs_exist_ok=True,
        )
        self._cleanup()

    def on_build_error(self, *, error: Exception) -> None:
        self._cleanup()


def is_notebook(*, relative_path: str | Path, notebook_patterns: Iterable[str]) -> bool:
    gi = gitmatch.compile(notebook_patterns, ignorecase=False)
    return bool(gi.match(Path(relative_path).as_posix()))


# Hooks for development
def on_startup(command: str, dirty: bool) -> None:
    log.info("[jupyterlite][development] plugin started.")


def on_page_markdown(markdown: str, page: Any, config: MkDocsConfig, files: Any) -> str:
    log.info("[jupyterlite][development] plugin started.")
    plugin = JupyterlitePlugin()
    return plugin.on_page_markdown(markdown, page=page, config=config, files=files)


def on_post_page(output: str, page: Page, config: MkDocsConfig) -> str:
    log.info("[jupyterlite][development] plugin started.")
    plugin = JupyterlitePlugin()
    return plugin.on_post_page(output, page=page, config=config)


def on_files(files: Files, config: MkDocsConfig) -> Files:
    log.info("[jupyterlite][development] plugin started.")
    plugin = JupyterlitePlugin()
    return plugin.on_files(files, config)


def get_nb_toc_and_title(path: str | Path) -> tuple[TableOfContents, str | None]:
    """Returns a TOC and title (the first heading, if present) for the Notebook."""
    notebook = nbformat.reads(Path(path).read_text(), as_version=4)
    (markdown_source, _resources) = MarkdownExporter().from_notebook_node(notebook)
    md = markdown.Markdown(extensions=["toc"])
    md.convert(markdown_source)
    toc = get_toc(md.toc_tokens)
    title = None
    for token in md.toc_tokens:
        if token["level"] == 1 and title is None:
            title = token["name"]
    return toc, title



from __future__ import annotations

import contextlib
import json
import logging
import os
import shutil
import subprocess
import tempfile
from collections.abc import Iterable
from pathlib import Path
from typing import Generator, Protocol
from urllib.parse import urlparse

log = logging.getLogger("mkdocs.plugins.jupyterlite")


class WheelSource(Protocol):
    command: str
    url: str


def build_site(
    *,
    docs_dir: Path,
    notebook_relative_paths: Iterable[Path],
    wheel_sources: Iterable[WheelSource],
    output_dir: Path,
) -> None:
    shutil.rmtree(output_dir, ignore_errors=True)
    output_dir.mkdir(parents=True, exist_ok=True)
    with _get_src_dir() as working_dir:
        log.debug(f"[jupyterlite] using working dir: {working_dir}")
        _write_jupyter_lite_json(working_dir)
        wheels_dir = working_dir / "wheels"
        wheel_urls = _get_wheel_urls(wheels_dir, wheel_sources)
        for notebook in notebook_relative_paths:
            src = docs_dir / notebook
            dst = working_dir / "files" / notebook
            dst.parent.mkdir(parents=True, exist_ok=True)
            log.debug(f"[jupyterlite] copying {src} to build {dst}")
            shutil.copy(src, dst)
        wheel_args = []
        for wheel_url in wheel_urls:
            wheel_args.extend(["--piplite-wheels", wheel_url])
        cmd = [
            "jupyter",
            "lite",
            "build",
            "--debug",
            "--contents",
            "files",
            *wheel_args,
            "--no-libarchive",
            "--apps",
            "notebooks",
            "--no-unused-shared-packages",
            "--output-dir",
            str(output_dir),
        ]
        log.info("[jupyterlite] running build command")
        _run_command(cmd, cwd=working_dir)
        assert output_dir.exists(), "Output directory was not created"


def _write_jupyter_lite_json(working_dir: Path) -> None:
    data = {
        "jupyter-lite-schema-version": 0,
        "jupyter-config-data": {
            # By default, jupyterlite saves the state of the notebook to the client's
            # browser, and on reload of the page, the notebook will be restored to that state.
            # The problem is that this local state overrides the contents sent from the server.
            # So, if you edit a notebook rebuild your docs, and refresh the page,
            # you still see the old version.
            # These settings make it so that the state is never stored on the client,
            # and is refreshed from the server on every page load.
            # Not ideal: it would be great if the user's state persisted until the
            # data on the server actually *changed*, but that doesn't appear possible yet.
            # See https://github.com/jupyterlite/jupyterlite/issues/1706#issuecomment-3187140714
            "enableMemoryStorage": True,
            "settingsStorageDrivers": ["memoryStorageDriver"],
            "contentsStorageDrivers": ["memoryStorageDriver"],
        },
    }
    path = working_dir / "jupyter-lite.json"
    path.write_text(json.dumps(data, indent=4))


def _run_command(cmd: list[str], **kwargs):
    cmd_str = cmd if isinstance(cmd, str) else " ".join(cmd)
    log.info("[jupyterlite] running command: " + cmd_str)
    try:
        result = subprocess.run(
            cmd,
            capture_output=True,
            check=True,
            text=True,
            **kwargs,
        )
        if result.stdout:
            log.debug("[jupyterlite] command output:\n" + result.stdout)
        if result.stderr:
            log.debug("[jupyterlite] command stderr:\n" + result.stderr)
    except subprocess.CalledProcessError as e:
        log.error("[jupyterlite] command failed with error code: " + str(e.returncode))
        if e.stdout:
            log.error("[jupyterlite] command stdout:\n" + e.stdout)
        if e.stderr:
            log.error("[jupyterlite] command stderr:\n" + e.stderr)
        raise


def _get_wheel_urls(
    working_dir: Path, wheel_sources: Iterable[WheelSource]
) -> list[str]:
    """Returns a list of wheel paths for the given wheel sources."""
    wheel_urls = []
    wheels_dir = working_dir.absolute() / "wheels"
    wheels_dir.mkdir(parents=True, exist_ok=True)
    for source in wheel_sources:
        if source.url and source.command:
            raise ValueError("Wheel source cannot have both URL and command", source)
        if not source.url and not source.command:
            raise ValueError("Wheel source must have either URL or command", source)
        if source.url:
            log.info(f"[jupyterlite] including wheel URL: {source.url}")
            url = urlparse(source.url)
            if not url.scheme or url.scheme == "file":
                # Local file URL
                wheel_file = Path(url.path).name
                dst = shutil.copy(source.url, wheels_dir / wheel_file)
                wheel_urls.append(str(dst.absolute()))
            else:
                wheel_urls.append(source.url)
        elif source.command:
            with tempfile.TemporaryDirectory() as tmpdir:
                actual_command = source.command.format(wheels_dir=tmpdir)
                log.info("[jupyterlite] running wheel command")
                _run_command(actual_command, shell=True)
                wheel_files = list(Path(tmpdir).glob("*.whl"))
                if not wheel_files:
                    # This is explicitly OK.
                    pass
                for wheel in wheel_files:
                    file_name = wheel.name
                    dst = shutil.copy(wheel, wheels_dir / file_name)
                    log.debug(f"[jupyterlite] including wheel URL: {file_name}")
                    wheel_urls.append(str(dst.absolute()))
    return wheel_urls


@contextlib.contextmanager
def _get_src_dir() -> Generator[Path, None, None]:
    # For debugging, allow setting the build directory
    if (src_dir_str := os.environ.get("MKDOCS_JUPYTERLITE_SRC_DIR")) is not None:
        p = Path(src_dir_str)
        shutil.rmtree(p, ignore_errors=True)
        p.mkdir(parents=True, exist_ok=True)
        yield p
    else:
        with tempfile.TemporaryDirectory() as tmpdir:
            yield Path(tmpdir)



from mkdocs_jupyterlite._plugin import is_notebook


def test_notebook_patterns():
    def check(path: str, patterns: list[str]) -> bool:
        # This redirection allows us to pass as positional args
        return is_notebook(relative_path=path, notebook_patterns=patterns)

    patterns = [
        "**/*.ipynb",  # include all
        "!**/draft_*.ipynb",  # drop drafts
        "/project/drafts/draft_keep.ipynb",  # re-include a specific draft
        "!/top_secret.ipynb",  # exclude anchored
    ]
    assert check("a/draft_temp.ipynb", patterns) is False
    assert check("project/drafts/draft_keep.ipynb", patterns) is True
    assert check("x/top_secret.ipynb", patterns) is True
    assert check("top_secret.ipynb", patterns) is False
